@@ -14,28 +14,45 @@
 #include "soup-auth.h"
 #include "soup-auth-basic.h"
 #include "soup-auth-digest.h"
+#include "soup-headers.h"
 
-#define PARENT_TYPE G_TYPE_OBJECT
-static GObjectClass *parent_class;
+G_DEFINE_TYPE (SoupAuth, soup_auth, G_TYPE_OBJECT)
 
 static void
-class_init (GObjectClass *object_class)
+finalize (GObject *object)
 {
-	parent_class = g_type_class_ref (PARENT_TYPE);
+	SoupAuth *auth = SOUP_AUTH (object);
+
+	g_free (auth->realm);
+
+	G_OBJECT_CLASS (soup_auth_parent_class)->finalize (object);
 }
 
-SOUP_MAKE_TYPE (soup_auth, SoupAuth, class_init, NULL, PARENT_TYPE)
+static void
+soup_auth_class_init (SoupAuthClass *auth_class)
+{
+	GObjectClass *object_class = G_OBJECT_CLASS (auth_class);
 
+	object_class->finalize = finalize;
+}
+
+static void
+soup_auth_init (SoupAuth *auth)
+{
+}
+
+typedef GType (*GTypeFunc) (void);
 
 typedef struct {
-	const char  *scheme;
-	GType      (*type_func) (void);
-	int          strength;
+	const char *name;
+	int         len;
+	GTypeFunc   type_func;
+	int         strength;
 } AuthScheme; 
 
 static AuthScheme known_auth_schemes [] = {
-	{ "Basic",  soup_auth_basic_get_type,  0 },
-	{ "Digest", soup_auth_digest_get_type, 3 },
+	{ "Basic",  sizeof ("Basic") - 1,  soup_auth_basic_get_type,  0 },
+	{ "Digest", sizeof ("Digest") - 1, soup_auth_digest_get_type, 3 },
 	{ NULL }
 };
 
@@ -53,18 +70,21 @@ static AuthScheme known_auth_schemes [] = {
 SoupAuth *
 soup_auth_new_from_header_list (const GSList *vals)
 {
-	char *header = NULL;
+	char *header = NULL, *realm;
 	AuthScheme *scheme = NULL, *iter;
 	SoupAuth *auth = NULL;
+	GHashTable *params;
 
 	g_return_val_if_fail (vals != NULL, NULL);
 
 	while (vals) {
 		char *tryheader = vals->data;
 
-		for (iter = known_auth_schemes; iter->scheme; iter++) {
-			if (!g_strncasecmp (tryheader, iter->scheme, 
-					    strlen (iter->scheme))) {
+		for (iter = known_auth_schemes; iter->name; iter++) {
+			if (!g_ascii_strncasecmp (tryheader, iter->name, 
+						  iter->len) &&
+			    (!tryheader[iter->len] ||
+			     g_ascii_isspace (tryheader[iter->len]))) {
 				if (!scheme || 
 				    scheme->strength < iter->strength) {
 					header = tryheader;
@@ -81,11 +101,20 @@ soup_auth_new_from_header_list (const GSList *vals)
 	if (!scheme)
 		return NULL;
 
-	auth = g_object_new (scheme->type_func (), NULL);
-	if (!auth)
+	params = soup_header_param_parse_list (header + scheme->len);
+	if (!params)
 		return NULL;
+	realm = soup_header_param_copy_token (params, "realm");
+	if (!realm) {
+		soup_header_param_destroy_hash (params);
+		return NULL;
+	}
 
-	SOUP_AUTH_GET_CLASS (auth)->construct (auth, header);
+	auth = g_object_new (scheme->type_func (), NULL);
+	auth->realm = realm;
+
+	SOUP_AUTH_GET_CLASS (auth)->construct (auth, params);
+	soup_header_param_destroy_hash (params);
 	return auth;
 }
 
@@ -96,7 +125,7 @@ soup_auth_new_from_header_list (const GSList *vals)
  * @password: the password provided by the user or client
  *
  * This is called by the session after requesting a username and
- * password from the application. @auth should take the information
+ * password from the application. @auth will take the information
  * and do whatever scheme-specific processing is needed.
  **/
 void
@@ -129,7 +158,7 @@ soup_auth_get_scheme_name (SoupAuth *auth)
  * soup_auth_get_realm:
  * @auth: a #SoupAuth
  *
- * Returns @auth's realm, if any.
+ * Returns @auth's realm.
  *
  * Return value: the realm name
  **/
@@ -138,7 +167,28 @@ soup_auth_get_realm (SoupAuth *auth)
 {
 	g_return_val_if_fail (SOUP_IS_AUTH (auth), NULL);
 
-	return SOUP_AUTH_GET_CLASS (auth)->get_realm (auth);
+	return auth->realm;
+}
+
+/**
+ * soup_auth_get_info:
+ * @auth: a #SoupAuth
+ *
+ * Gets an identifier for @auth. #SoupAuth objects from the same
+ * server with the same identifier refer to the same authentication
+ * domain (eg, the URLs associated with them take the same usernames
+ * and passwords).
+ *
+ * Return value: the identifier
+ **/
+char *
+soup_auth_get_info (SoupAuth *auth)
+{
+	g_return_val_if_fail (SOUP_IS_AUTH (auth), NULL);
+
+	return g_strdup_printf ("%s:%s",
+				SOUP_AUTH_GET_CLASS (auth)->scheme_name,
+				auth->realm);
 }
 
 /**
@@ -162,8 +212,9 @@ soup_auth_is_authenticated (SoupAuth *auth)
  * @auth: a #SoupAuth
  * @msg: the #SoupMessage to be authorized
  *
- * Generates an appropriate "Authorization" header for @msg. (This
- * will only be called if soup_auth_is_authenticated() returns %TRUE.)
+ * Generates an appropriate "Authorization" header for @msg. (The
+ * session will only call this if soup_auth_is_authenticated()
+ * returned %TRUE.)
  *
  * Return value: the "Authorization" header, which must be freed.
  **/
