@@ -7,6 +7,9 @@
 #define SOUP_DATE_H 1
 
 #include <time.h>
+#include <glib/gmacros.h>
+
+G_BEGIN_DECLS
 
 time_t  soup_mktime_utc         (struct tm *tm);
 void    soup_gmtime             (const time_t *when, struct tm *tm);
@@ -16,4 +19,6 @@ char   *soup_date_generate      (time_t when);
 
 time_t  soup_date_iso8601_parse (const char *timestamp);
 
+G_END_DECLS
+
 #endif /* SOUP_DATE_H */
