@@ -5,11 +5,26 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
+#include <unistd.h>
 
 #include "libsoup/soup.h"
 #include "apache-wrapper.h"
 
 int errors = 0;
+gboolean debug = FALSE;
+
+static void
+dprintf (const char *format, ...)
+{
+	va_list args;
+
+	if (!debug)
+		return;
+
+	va_start (args, format);
+	vprintf (format, args);
+	va_end (args);
+}
 
 typedef struct {
 	const char *explanation;
@@ -61,7 +76,7 @@ test_url (const char *url, int proxy, guint expected, gboolean sync)
 	SoupUri *proxy_uri;
 	SoupMessage *msg;
 
-	printf ("  GET %s via %s\n", url, proxy_names[proxy]);
+	dprintf ("  GET %s via %s\n", url, proxy_names[proxy]);
 	if (proxy == UNAUTH_PROXY && expected != SOUP_STATUS_FORBIDDEN)
 		expected = SOUP_STATUS_PROXY_UNAUTHORIZED;
 
@@ -84,9 +99,9 @@ test_url (const char *url, int proxy, guint expected, gboolean sync)
 
 	soup_session_send_message (session, msg);
 
-	printf ("  %d %s\n", msg->status_code, msg->reason_phrase);
+	dprintf ("  %d %s\n", msg->status_code, msg->reason_phrase);
 	if (msg->status_code != expected) {
-		printf ("  EXPECTED %d!\n", expected);
+		dprintf ("  EXPECTED %d!\n", expected);
 		errors++;
 	}
 
@@ -100,8 +115,8 @@ run_test (int i, gboolean sync)
 {
 	char *http_url, *https_url;
 
-	printf ("Test %d: %s (%s)\n", i + 1, tests[i].explanation,
-		sync ? "sync" : "async");
+	dprintf ("Test %d: %s (%s)\n", i + 1, tests[i].explanation,
+		 sync ? "sync" : "async");
 
 	if (!strncmp (tests[i].url, "http", 4)) {
 		http_url = g_strdup (tests[i].url);
@@ -111,26 +126,43 @@ run_test (int i, gboolean sync)
 		https_url = g_strconcat (HTTPS_SERVER, tests[i].url, NULL);
 	}
 	test_url (http_url, SIMPLE_PROXY, tests[i].final_status, sync);
+#if HAVE_SSL
 	test_url (https_url, SIMPLE_PROXY, tests[i].final_status, sync);
+#endif
 	test_url (http_url, AUTH_PROXY, tests[i].final_status, sync);
+#if HAVE_SSL
 	test_url (https_url, AUTH_PROXY, tests[i].final_status, sync);
+#endif
 	test_url (http_url, UNAUTH_PROXY, tests[i].final_status, sync);
+#if HAVE_SSL
 	test_url (https_url, UNAUTH_PROXY, tests[i].final_status, sync);
+#endif
 
 	g_free (http_url);
 	g_free (https_url);
 
-	printf ("\n");
+	dprintf ("\n");
 }
 
 int
 main (int argc, char **argv)
 {
-	int i;
+	int i, opt;
 
 	g_type_init ();
 	g_thread_init (NULL);
 
+	while ((opt = getopt (argc, argv, "d")) != -1) {
+		switch (opt) {
+		case 'd':
+			debug = TRUE;
+			break;
+		default:
+			fprintf (stderr, "Usage: %s [-d]\n", argv[0]);
+			return 1;
+		}
+	}
+
 	if (!apache_init ()) {
 		fprintf (stderr, "Could not start apache\n");
 		return 1;
@@ -142,7 +174,13 @@ main (int argc, char **argv)
 	}
 
 	apache_cleanup ();
+	g_main_context_unref (g_main_context_default ());
 
-	printf ("proxy-test: %d errors\n", errors);
+	dprintf ("\n");
+	if (errors) {
+		printf ("proxy-test: %d error(s). Run with '-d' for details\n",
+			errors);
+	} else
+		printf ("proxy-test: OK\n");
 	return errors;
 }
